import logging


def install_colored_logs(logger: logging.Logger, level: int = None):
    import coloredlogs
    # NOTE: This is kinda weird, but it seems that the coloredlogs library has
    #       some issues: https://github.com/xolox/python-coloredlogs/issues/18
    # get the current root logger level
    logLevel = logging.getLogger().getEffectiveLevel()
    # install the coloredlogs on our logger and with our level
    coloredlogs.install(level=level or logger.level, logger=logger)
    # restore the root logger level
    logging.getLogger().setLevel(logLevel)



from dt_shell import DTCommandAbs, DTShell
from dt_shell.constants import EMBEDDED_COMMAND_SET_NAME
from typing import List


class DTCommand(DTCommandAbs):
    help = "Shows the list of all the commands available in the shell."

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # show core commands
        print("\nCore commands:")
        keys = sorted(shell.command_set(EMBEDDED_COMMAND_SET_NAME).commands.keys())
        length = len(max(keys, key=len)) + 2
        command_descriptions = shell.profile.command_descriptions
        for cmd in keys:
            print("\t%-*s%s" % (length, cmd, command_descriptions[cmd]["description"] if cmd in command_descriptions else ""))
        # show commands grouped by command sets
        for cs in shell.command_sets:
            if cs.name == EMBEDDED_COMMAND_SET_NAME:
                continue
            print(f"\nCommand set '{cs.name}':")
            keys = sorted(cs.commands.keys())
            length = len(max(keys, key=len)) + 2
            for cmd in keys:
                print("\t%-*s%s" % (length, cmd, command_descriptions[cmd]["description"] if cmd in command_descriptions else ""))

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



# This replaces the old default __init__ file for the Duckietown Shell commands
#
# Maintainer: Andrea F. Daniele
import copy
import importlib
import importlib.util
import os.path
import sys
from os.path import (
    exists as _exists,
    dirname as _dirname,
    relpath as _relpath,
    join as _join,
)
from typing import Type, List

from dt_shell.constants import DTShellConstants

from .. import logger
from ..exceptions import ShellNeedsUpdate, CommandsLoadingException
from .commands import default_command_configuration, failed_to_load_command, DTCommandConfigurationAbs, \
    DTCommandAbs, CommandSet, DTCommandSetConfigurationAbs, default_commandset_configuration


def import_commandset_configuration(command_set: CommandSet) -> Type[DTCommandSetConfigurationAbs]:
    # constants
    _configuration_file = _join(command_set.path, "__command_set__", "configuration.py")
    # import command set configuration
    if _exists(_configuration_file):
        if DTShellConstants.VERBOSE:
            logger.debug(f"Importing configuration for command set '{command_set.name}' from "
                         f"'{_configuration_file}'")
        # use a unique module name to avoid caching issues between command sets
        _module_name: str = f"__command_set_config_{command_set.name}__"
        # we temporarily add the path to the command set to PYTHONPATH
        old: List[str] = copy.deepcopy(sys.path)
        sys.path.insert(0, os.path.abspath(command_set.path))
        try:
            # load the configuration module from the specific file
            spec = importlib.util.spec_from_file_location(_module_name, _configuration_file)
            if spec is None or spec.loader is None:
                msg = f"Cannot load command set configuration module from {_configuration_file}"
                raise CommandsLoadingException(msg)
            configuration = importlib.util.module_from_spec(spec)
            sys.modules[_module_name] = configuration
            spec.loader.exec_module(configuration)
        finally:
            # restore PYTHONPATH
            sys.path = old

        DTCommandSetConfiguration: Type[DTCommandSetConfigurationAbs] = \
            configuration.DTCommandSetConfiguration
        if not issubclass(DTCommandSetConfiguration.__class__, DTCommandSetConfigurationAbs.__class__):
            msg = f"Cannot load command set configuration class in {_configuration_file}, the class " \
                  f"'DTCommandSetConfiguration' must extend the class 'DTCommandSetConfigurationAbs'"
            raise CommandsLoadingException(msg)
        # populate path
        DTCommandSetConfiguration.path = command_set.path

        return DTCommandSetConfiguration
    else:
        # populate path on default configuration as well
        default_commandset_configuration.DTCommandSetConfiguration.path = command_set.path
        return default_commandset_configuration.DTCommandSetConfiguration


def import_commandset_init(command_set: CommandSet):
    # constants
    _init_file = _join(command_set.path, "__command_set__", "init.py")
    # skip if there is no init file for this command set
    if not _exists(_init_file):
        return
    # import command set init
    if DTShellConstants.VERBOSE:
        logger.debug(f"Executing init script for command set '{command_set.name}' from '{_init_file}'")
    _init_sel: str = "__command_set__.init"
    # we temporarily add the path to the command set to PYTHONPATH
    old: List[str] = copy.deepcopy(sys.path)
    sys.path.insert(0, os.path.abspath(command_set.path))
    # import/refresh the init module
    try:
        init = importlib.import_module(_init_sel)
        if init.__file__ != _init_file:
            importlib.reload(init)
        # make sure we got the right one
        assert init.__file__ == _init_file
    finally:
        # restore PYTHONPATH
        sys.path = old


def import_configuration(command_set: CommandSet, selector: str) -> Type[DTCommandConfigurationAbs]:
    # constants
    _command_dir = command_set.command_path(selector)
    _configuration_file = _join(_command_dir, "configuration.py")
    # import command configuration
    if _exists(_configuration_file):
        _command_sel: str = _relpath(_command_dir, command_set.path).strip("/").replace("/", ".")
        _configuration_sel: str = f"{_command_sel}.configuration"
        # we temporarily add the path to the command set to PYTHONPATH
        old: List[str] = copy.deepcopy(sys.path)
        sys.path.insert(0, os.path.abspath(command_set.path))
        try:
            if DTShellConstants.VERBOSE:
                logger.debug(f"Importing configuration for command '{_command_sel}' from "
                             f"'{_configuration_file}'")
            configuration = importlib.import_module(_configuration_sel)
        except ShellNeedsUpdate as e:
            logger.warning(
                f"Command '{_command_sel}' was not loaded because the shell needs to be "
                f"updated. Current version is {e.current_version}, required version is "
                f"{e.version_needed}"
            )
            return default_command_configuration.DTCommandConfiguration
        except ModuleNotFoundError as e:
            logger.warning(
                f"Command configuration '{_command_sel}' could not be loaded due to missing dependency: {e}. "
                f"Using default configuration."
            )
            return default_command_configuration.DTCommandConfiguration
        except Exception as e:
            logger.warning(
                f"Command configuration '{_command_sel}' could not be loaded: {e}. "
                f"Using default configuration."
            )
            return default_command_configuration.DTCommandConfiguration
        finally:
            # restore PYTHONPATH
            sys.path = old

        DTCommandConfiguration = configuration.DTCommandConfiguration
        if not issubclass(DTCommandConfiguration.__class__, DTCommandConfigurationAbs.__class__):
            msg = f"Cannot load command configuration class in {_configuration_file}, the class " \
                  f"'DTCommandConfiguration' must extend the class 'DTCommandConfigurationAbs'"
            raise CommandsLoadingException(msg)

        return DTCommandConfiguration
    else:
        return default_command_configuration.DTCommandConfiguration


def import_command(command_set: CommandSet, fpath: str) -> Type[DTCommandAbs]:
    # we always start from an __init__.py file
    if os.path.isdir(fpath):
        fpath = os.path.join(fpath, "__init__.py")
    # constants
    _command_dir = _dirname(fpath) if os.path.isfile(fpath) else fpath
    _command_file = _join(_command_dir, "command.py")
    # import current command
    if _exists(_command_file):
        _command_sel: str = _relpath(_command_dir, command_set.path).strip("/").replace("/", ".")
        _command_sel: str = f"{_command_sel}.command"
        try:
            if DTShellConstants.VERBOSE:
                logger.debug(f"Importing command '{_command_sel}' from '{_dirname(fpath)}/'")

            # when running in debug mode (via vscode debugpy), it seems that Python's `code.py` is loaded
            # and it clashes with the `code` commands... The code below checks for such clashes, and
            # unloads the module if not part of the commandset.

            # see implementation of `_find_and_load_unlocked` in importlib._bootstrap
            _command_sel_parent = _command_sel.split('.')[0]
            if _command_sel_parent and _command_sel_parent in sys.modules:
                mod = sys.modules.get(_command_sel_parent, None)
                # if (_command_sel_parent == "code"):
                #     print(dir(mod))
                #     print(mod.__path__)
                if "__path__" not in dir(mod) or mod.__path__[0] not in _command_dir:
                    # print(f"{_command_dir} == {mod.__path__[0]}")
                    logger.warning(
                        f"Command '{_command_sel}' is already loaded or there's a name "
                        f"clash with '{_command_sel_parent}:{mod.__file__}'. Unloading '{_command_sel_parent}' "
                        f"to load '{_command_sel}'"
                    )
                    # see https://stackoverflow.com/questions/43181440/what-does-del-sys-modulesmodule-actually-do
                    del sys.modules[_command_sel_parent]
            command = importlib.import_module(_command_sel)
        except ShellNeedsUpdate as e:
            logger.warning(
                f"Command '{_command_sel}' was not loaded because the shell needs to be "
                f"updated. Current version is {e.current_version}, required version is "
                f"{e.version_needed}"
            )
            return failed_to_load_command.DTCommand

        DTCommand = command.DTCommand        # handle wrong class
        if not issubclass(DTCommand.__class__, DTCommandAbs.__class__):
            msg = f"Cannot load command class in {_command_file}, the class " \
                  f"'DTCommand' must extend the class 'DTCommandAbs'"
            raise CommandsLoadingException(msg)
        return DTCommand
    else:
        return failed_to_load_command.DTCommand



import os.path
import shutil
from typing import List

import questionary

from dt_shell import DTCommandAbs, DTShell, dtslogger


class DTCommand(DTCommandAbs):
    help = "Resets the current profile Python's virtual environment"

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        # make sure the user known what is doing
        dtslogger.warning(
            "\n"
            "---\n"
            "\n"
            "This operation will delete this profile's virtual environment.\n"
            "This is usually a safe operation as the environment will be recreated the next time the shell is run.\n"
            "\n---"
        )
        proceed: bool = questionary.confirm("Do you want to continue?").ask()
        # ---
        if not proceed:
            dtslogger.info("Operation aborted!")
            return

        # delete virtual environment
        venv_path: str = os.path.join(shell.profile.path, "venv")
        if not os.path.exists(venv_path):
            dtslogger.error(f"The virtual environment was expected to be found at '{venv_path}' but this path "
                            f"does not exist. This should not have happened.")
            return
        dtslogger.info("Deleting virtual environment...")
        dtslogger.debug(f"Removing path '{venv_path}'")
        shutil.rmtree(venv_path)
        dtslogger.info("Virtual environment successfully reset.")

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



import logging
import warnings

from dt_shell_cli.utils import install_colored_logs

warnings.filterwarnings(action='ignore', module='.*paramiko.*')

logging.basicConfig()

# logger dedicated to the shell
logger = logging.getLogger("shell")

# add colored logs
install_colored_logs(logger=logger)
logger.setLevel(logging.INFO)



import argparse
from typing import Optional, List

from dt_shell.commands import DTCommandConfigurationAbs
from dt_shell.environments import ShellCommandEnvironmentAbs


class DTCommandConfiguration(DTCommandConfigurationAbs):

    @classmethod
    def environment(cls, *args, **kwargs) -> Optional[ShellCommandEnvironmentAbs]:
        """
        The environment in which this command will run.
        """
        return None

    @classmethod
    def parser(cls, *args, **kwargs) -> Optional[argparse.ArgumentParser]:
        """
        The parser this command will use.
        """
        parser = argparse.ArgumentParser()
        parser.add_argument('repository', help='The repository containing the command set to install. '
                                               'Format: "[server/]owner/repo" or a fully qualified URL.')
        parser.add_argument('branch', nargs='?', default='main', help='The branch to install. (default: main)')
        return parser

    @classmethod
    def aliases(cls) -> List[str]:
        """
        Alternative names for this command.
        """
        return []



import argparse
import os
from typing import List, cast, Optional

import requests

from dt_shell import DTCommandAbs, DTShell, dtslogger, RunCommandException
from dt_shell.commands import CommandSet
from dt_shell.commands.repository import CommandsRepository
from dt_shell.constants import DB_USER_COMMAND_SETS_REPOSITORIES
from dt_shell.database import DTShellDatabase
from dt_shell.utils import run_cmd


class DTCommand(DTCommandAbs):
    help = "Installs a new command set."

    @staticmethod
    def command(shell: DTShell, args: List[str]):
        parsed: argparse.Namespace = DTCommand.parser.parse_args(args)
        # ---
        url: Optional[str] = None
        # sanitize the repository URL
        parsed.repository = parsed.repository.strip("/")
        if parsed.repository.endswith(".git"):
            parsed.repository = parsed.repository[:-4]
        # - public URL
        if ":" not in parsed.repository:
            dtslogger.debug(f"Checking if '{parsed.repository}' is a public repository.")
            # assume the URL is fully qualified
            public_url = parsed.repository
            # add the protocol if missing
            if cast(str, parsed.repository).count("/") == 2:
                # matches the pattern "server/owner/repo"
                public_url = f"https://{public_url}"
            # default to github.com
            if not public_url.startswith('http'):
                public_url = f"https://github.com/{parsed.repository}"
            # ---
            url = public_url
            dtslogger.debug(f"Assuming the URL is the following: {url}")
            # try to reach the repository at the public URL
            repo: CommandsRepository = CommandsRepository.from_remoteurl(url, parsed.branch)
            try:
                dtslogger.debug(f"> HEAD {repo.apiurl}")
                requests.head(repo.apiurl, timeout=5).raise_for_status()
                dtslogger.info(f"Repository found at {public_url}.")
            except requests.RequestException:
                dtslogger.info(f"Repository not found at {public_url}. Assuming it is a private repository.")
                url = None

        if ":" in parsed.repository or url is None:
            # assume the URL is a private repository
            dtslogger.debug(f"Assuming '{parsed.repository}' is a private repository.")
            # - private URL
            private_url = parsed.repository
            # add the protocol if missing
            if ":" in parsed.repository:
                if not private_url.startswith('git@'):
                    private_url = f"git@{private_url}"
            else:
                private_url = f"git@github.com:{parsed.repository}"
            # ---
            url = private_url
            dtslogger.debug(f"Assuming the URL is the following: {url}")
            # check if SSH is configured properly
            dtslogger.debug("Checking if SSH is configured properly.")
            try:
                cmd = ["ssh", "-T", "git@github.com"]
                dtslogger.debug(f"$ {' '.join(cmd)}")
                run_cmd(cmd)
            except RunCommandException as e:
                if e.exit_code == 1 and "successfully" in e.stderr.lower():
                    dtslogger.info("SSH communication with GitHub is successful.")
                else:
                    dtslogger.error("SSH is not configured properly. The command 'ssh -T git@github.com' fails.")
                    return
        # ---
        # define command set properties
        repository: CommandsRepository = CommandsRepository.from_remoteurl(url, parsed.branch)
        name: str = f"{repository.username}__{repository.project}"
        path: str = os.path.join(shell.profile.path, "commands", name)
        # create command set
        cs: CommandSet = CommandSet(
            name=name,
            path=path,
            profile=shell.profile,
            repository=repository,
        )
        # open the database
        db: DTShellDatabase[dict] = shell.profile.database(DB_USER_COMMAND_SETS_REPOSITORIES)
        # make sure the command set is not already installed
        if db.contains(cs.name):
            dtslogger.info(f"Command set '{cs.name}' is already installed.")
            return
        # install command set
        dtslogger.info(f"Downloading command set '{cs.name}'...")
        cs.ensure_commands_exist()
        # add commands to the list of user command sets
        dtslogger.debug(f"Adding command set '{cs.name}' to the list of user command sets.")
        db.set(cs.name, cs.repository.as_dict())
        # done
        dtslogger.info(f"Command set '{cs.name}' installed successfully.")

    @staticmethod
    def complete(shell: DTShell, word: str, line: str) -> List[str]:
        return []



import json
import os
import traceback
from dataclasses import dataclass
from typing import Optional, List

from dt_shell_cli import logger
from ..exceptions import RunCommandException
from ..checks.version import get_url
from ..constants import DEFAULT_COMMAND_SET_REPOSITORY, DTShellConstants
from ..utils import run_cmd, provider_username_project_from_git_url, indent_block


@dataclass
class CommandsRepository:
    username: str
    project: str
    branch: str
    location: Optional[str] = None
    use_ssh: bool = False
    provider: str = "github.com"

    @property
    def remoteurl(self) -> str:
        if self.use_ssh:
            return f"git@{self.provider}:{self.username}/{self.project}"
        return f"https://{self.provider}/{self.username}/{self.project}"

    @property
    def apiurl(self) -> str:
        if self.provider != "github.com":
            raise ValueError("Only github.com is supported")
        return f"https://api.github.com/repos/{self.username}/{self.project}"

    @property
    def apiurl_with_branch(self) -> str:
        if self.provider != "github.com":
            raise ValueError("Only github.com is supported")
        return f"https://api.github.com/repos/{self.username}/{self.project}/branches/{self.branch}"

    def as_dict(self) -> dict:
        return {
            "provider": self.provider,
            "username": self.username,
            "project": self.project,
            "branch": self.branch,
            "location": self.location,
            "url": self.remoteurl,
        }

    @classmethod
    def from_dict(cls, data: dict) -> 'CommandsRepository':
        return CommandsRepository(
            username=data["username"],
            project=data["project"],
            branch=data["branch"],
            location=data.get("location", None),
            use_ssh=data["url"].startswith("git@"),
            provider=data.get("provider", "github.com")
        )

    def remote_sha(self) -> Optional[str]:
        # Get the remote sha from GitHub
        logger.info("Fetching remote SHA from github.com ...")
        if self.use_ssh:
            logger.warning(f"Command set '{self.username}/{self.project}' is using SSH. Cannot fetch remote SHA.")
            return None
        else:
            remote_url: str = self.apiurl_with_branch
            # contact github
            try:
                content = get_url(remote_url)
            except Exception:
                if DTShellConstants.VERBOSE:
                    traceback.print_exc()
                logger.debug(f"URL called: {remote_url}")
                logger.warning(f"An error occurred while fetching the remote SHA for repository {self.remoteurl}")
                return None
            # parse output
            try:
                data = json.loads(content)
                return data["commit"]["sha"]
            except Exception:
                if DTShellConstants.VERBOSE:
                    traceback.print_exc()
                logger.debug(f"URL called: {remote_url}\n"
                             f"Object returned by github:\n\n{indent_block(content)}")
                logger.warning(f"An error occurred while fetching the remote SHA for repository {self.remoteurl}")
                return None

    @classmethod
    def given_distro(cls, distro: str) -> 'CommandsRepository':
        return CommandsRepository(
            **DEFAULT_COMMAND_SET_REPOSITORY,
            branch=distro
        )

    @classmethod
    def from_file_system(cls, path: str, location: Optional[str] = None) -> 'CommandsRepository':
        origin_url = run_cmd(["git", "-C", path, "config", "--get", "remote.origin.url"])
        _, username, project = provider_username_project_from_git_url(origin_url)
        branch = run_cmd(["git", "-C", path, "rev-parse", "--abbrev-ref", "HEAD"])
        return CommandsRepository(username, project, branch, location=location)

    @staticmethod
    def head_tag(path: str) -> Optional[str]:
        # Check if path exists first
        if not os.path.exists(path):
            return None
        try:
            tags_output: Optional[str] = run_cmd(["git", "-C", path, "describe", "--exact-match", "--tags", "HEAD"])
        except RunCommandException as e:
            if e.stderr is not None and "no tag exactly matches" in e.stderr:
                # just not an exact match
                return None
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            return None
        except Exception:
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            return None
        # ---
        if tags_output is None:
            return None
        tags: List[str] = tags_output.strip().split("\n")
        if not tags:
            return None
        return tags[0]

    @staticmethod
    def closest_tag(path: str) -> Optional[str]:
        # Check if path exists first
        if not os.path.exists(path):
            return None
        try:
            tags_output: Optional[str] = run_cmd(["git", "-C", path, "tag", "--merged"])
        except Exception:
            if DTShellConstants.VERBOSE:
                traceback.print_exc()
            return None
        # no tags
        if tags_output is None:
            return None
        # ---
        tags: List[str] = tags_output.strip().split("\n")
        if not tags:
            return None
        return tags[-1]

    @classmethod
    def from_remoteurl(cls, remoteurl: str, branch: str, location: Optional[str] = None) -> 'CommandsRepository':
        provider, username, project = provider_username_project_from_git_url(remoteurl)
        return CommandsRepository(
            username=username,
            project=project,
            branch=branch,
            provider=provider,
            location=location,
            use_ssh=remoteurl.startswith("git@")
        )



import logging
import os
import sys
import platform
import re
import shutil
import subprocess
import importlib
import locale
import json
import yaml
import traceback
from math import floor
from traceback import format_exc
from typing import Optional, Any, Tuple, List, Dict, Union

import termcolor
from termcolor import colored
from questionary import Style

import dt_authentication
from dt_authentication import DuckietownToken

from dt_shell_cli import logger
from . import __version__
from .constants import BASH_COMPLETION_DIR, SHELL_LIB_DIR, DTShellConstants
from .exceptions import ShellInitException, RunCommandException

NOTSET = object()
MAX_PIP_INSTALL_ATTEMPTS = 2


cli_style = Style([
    ('qmark', 'fg:#673ab7 bold'),        # token in front of the question
    ('question', 'bold'),                # question text
    ('choice', 'fg:#fec20b bold'),       # a possible choice in select
    ('answer', 'fg:#fec20b bold'),       # submitted answer text behind the question
    ('pointer', 'fg:#673ab7 bold'),      # pointer used in select and checkbox prompts
    ('highlighted', 'fg:#673ab7 bold'),  # pointed-at choice in select and checkbox prompts
    ('selected', 'fg:#cc5454'),          # style for a selected item of a checkbox
    ('separator', 'fg:#cc5454'),         # separator in lists
    ('instruction', ''),                 # user instructions for select, rawselect, checkbox
    ('text', ''),                        # plain text
    ('disabled', 'fg:#bbbbbb italic')    # disabled choices for select and checkbox prompts
])


def indent(s: str, prefix: str, first: Optional[str] = None) -> str:
    s = str(s)
    assert isinstance(prefix, str)
    lines = s.split("\n")
    if not lines:
        return ""

    if first is None:
        first = prefix

    m = max(len(prefix), len(first))

    prefix = " " * (m - len(prefix)) + prefix
    first = " " * (m - len(first)) + first

    # different first prefix
    res = ["%s%s" % (prefix, line.rstrip()) for line in lines]
    res[0] = "%s%s" % (first, lines[0].rstrip())
    return "\n".join(res)


def raise_wrapped(etype, e, msg, compact=False, exc=None, **kwargs):
    """Raises an exception of type etype by wrapping
    another exception "e" with its backtrace and adding
    the objects in kwargs as formatted by format_obs.

    if compact = False, write the whole traceback, otherwise just str(e).

    exc = output of sys.exc_info()
    """

    e = raise_wrapped_make(etype, e, msg, compact=compact, **kwargs)

    #     if exc is not None:
    #         _, _, trace = exc
    #         raise etype, e.args, trace
    #     else:
    raise e


def raise_wrapped_make(etype, e, msg, compact=False, **kwargs):
    """Constructs the exception to be thrown by raise_wrapped()"""
    assert isinstance(e, BaseException), type(e)
    assert isinstance(msg, str), type(msg)
    s = msg

    import sys

    if sys.version_info[0] >= 3:
        es = str(e)
    else:
        if compact:
            es = str(e)
        else:
            es = format_exc()

    s += "\n" + indent(es.strip(), "| ")

    return etype(s)


def check_isinstance(ob, expected, **kwargs):
    if not isinstance(ob, expected):
        kwargs["object"] = ob
        raise_type_mismatch(ob, expected, **kwargs)


def raise_type_mismatch(ob, expected, **kwargs):
    """Raises an exception concerning ob having the wrong type."""
    e = "Object not of expected type:"
    e += "\n  expected: %s" % str(expected)
    e += "\n  obtained: %s" % str(type(ob))
    # e += '\n' + indent(format_obs(kwargs), ' ')
    raise ValueError(e)


SPACE_TAG = "SPACE_TAG"


def replace_spaces(x: str) -> str:
    return x.replace(" ", SPACE_TAG)


def undo_replace_spaces(x: str) -> str:
    return x.replace(SPACE_TAG, " ")


def href(x):
    return termcolor.colored(x, "blue", None, ["underline"])


def dark_yellow(x):
    return termcolor.colored(x, "yellow")


def dark(x):
    return termcolor.colored(x, attrs=["dark"])


def safe_pathname(s: str) -> str:
    return re.sub(r"[^\w\d-]", "_", s)


def run_cmd(cmd, print_output=False, suppress_errors=False):
    logger.debug("$ %s" % cmd)
    # spawn new process
    proc = subprocess.Popen(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    stdout = stdout.decode("utf-8") if stdout else None
    stderr = stderr.decode("utf-8") if stderr else None
    returncode = proc.returncode
    # ---
    if returncode != 0 and not suppress_errors:
        msg = "The command %r failed with exit code %d.\nError:\n%s\nOutput:\n%s\n" % (
            cmd, returncode, indent_block(stderr), indent_block(stdout))
        raise RunCommandException(msg, returncode, stdout, stderr)
    if print_output:
        print(stdout)
    return stdout


def parse_version(x: str):
    return tuple(int(_) for _ in x.split("."))


def render_version(t: Tuple[int, int, int]):
    return ".".join(str(_) for _ in t)


def load_class(name: str, module: str, package: Optional[str] = None) -> Any:
    logger.debug(f"Loading class {package or ''}.{module}.{name}")
    mod = importlib.import_module(name=module, package=package)
    return getattr(mod, name)


def provider_username_project_from_git_url(url: str) -> Tuple[str, str, str]:
    ssh_pattern = "git@([^:]+):([^/]+)/(.+)"
    res = re.search(ssh_pattern, url, re.IGNORECASE)
    if res:
        return res.group(1), res.group(2), res.group(3)
    https_pattern = "https://([^/]+)/([^/]+)/(.+)"
    res = re.search(https_pattern, url, re.IGNORECASE)
    if res:
        return res.group(1), res.group(2), res.group(3)


def ln(s: str) -> int:
    """
    Computes the length of a string while taking into account the coloring characters.
    """
    return len(re.sub(rb"(\x1b|\[\d{1,2}m)*", b"", s.encode("utf-8")))


def text_distribute(chunks: List[str], width: int) -> str:
    remaining: int = width - sum(map(ln, chunks))
    spaces: List[int] = [0] + [int(floor(remaining / (len(chunks) - 1)))] * (len(chunks) - 1)
    spaces[-1] += remaining - sum(spaces)

    txt: str = ""
    for s, t in zip(spaces, chunks):
        txt += f"{' ' * s}{t}"
    return txt


def text_justify(txt: str, width: int) -> str:
    return "\n".join(
        [" " * int(floor((width - ln(line)) / 2)) + line for line in txt.splitlines(keepends=False)]
    )


def validator_token(token: str) -> bool:
    try:
        DuckietownToken.from_string(token, allow_expired=False)
    except dt_authentication.exceptions.InvalidToken as e:
        # logger.error(f"The given token is not valid: {str(e)}")
        return False
    except dt_authentication.exceptions.ExpiredToken:
        # logger.warning(f"The given token is expired, make sure you get a fresh one before continuing.")
        return False
    # ---
    return True


def yellow_bold(x: Any) -> str:
    return colored(str(x), color="yellow", attrs=["bold"])


class DebugInfo:
    name2versions: Dict[str, Union[str, Dict[str, str]]] = {}


def install_pip_tool(interpreter: str):
    get_pip_fpath: str = os.path.join(SHELL_LIB_DIR, "assets", "get-pip.py")
    if not os.path.exists(get_pip_fpath):
        msg = f"Required file for pip installation not found: {get_pip_fpath}"
        raise ShellInitException(msg)
    logger.info("Installing pip...")
    try:
        subprocess.check_output([interpreter, get_pip_fpath], stderr=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        # TODO: test this failure case on purpose
        msg: str = "An error occurred while installing pip in the virtual environment"
        raise ShellInitException(msg, stdout=e.stdout, stderr=e.stderr)


def pip_install(interpreter: str, requirements: str):
    run = subprocess.check_call if logger.level <= logging.DEBUG else subprocess.check_output
    for attempt in range(MAX_PIP_INSTALL_ATTEMPTS):
        try:
            run(
                [interpreter, "-m", "pip", "install", "-r", requirements],
                stderr=subprocess.STDOUT,
                env={}
            )
            break
        except subprocess.CalledProcessError as e:
            error = e.stdout.decode("utf-8", errors="replace") if e.stdout else ""
            if attempt == MAX_PIP_INSTALL_ATTEMPTS - 1 or "No module named pip" not in error:
                msg: str = "An error occurred while installing python dependencies"
                raise ShellInitException(msg, stdout=e.stdout, stderr=e.stderr)
            install_pip_tool(interpreter)


def indent_block(s: str, indent_len: int = 4) -> str:
    space: str = " " * indent_len
    return space + f"\n{space}".join(s.splitlines() if s is not None else ["None"])


def pretty_json(data: Any, indent_len: int = 0) -> str:
    return indent_block(json.dumps(data, sort_keys=True, indent=4), indent_len=indent_len)


def pretty_exc(exc: Exception, indent_len: int = 0) -> str:
    return indent_block(
        ''.join(traceback.TracebackException.from_exception(exc).format()), indent_len=indent_len)


def ensure_bash_completion_installed():
    import dt_shell
    from dt_shell.database import DTShellDatabase
    if platform.system() in ["Linux", "Darwin"]:
        db: DTShellDatabase = DTShellDatabase.open("bash-completion-install")
        key: str = f"dts-comletion-{dt_shell.__version__}"
        if not db.contains(key):
            logger.info("Installing bash-completion script...")
            src: str = os.path.join(SHELL_LIB_DIR, "assets", "dts-completion.bash")
            dst: str = os.path.join(BASH_COMPLETION_DIR, "dts")
            try:
                os.makedirs(BASH_COMPLETION_DIR, exist_ok=True)
                shutil.copyfile(src, dst)
            except Exception:
                logger.warning("An error occurred while attempting to install the bash-completion script")
                if DTShellConstants.VERBOSE:
                    traceback.print_last()
                return
            # mark it as installed in the database
            db.set(key, True)
            logger.info("Bash-completion script successfully installed")
        else:
            logger.debug("Bash-completion script marked as already installed")


def print_debug_info() -> None:
    from .logging import dts_print
    from .checks.packages import _get_installed_distributions

    v = DebugInfo.name2versions
    v["python"] = sys.version
    v["duckietown-shell"] = __version__

    v["encodings"] = {
        "stdout": sys.stdout.encoding,
        "stderr": sys.stderr.encoding,
        "locale": locale.getpreferredencoding(),
    }

    try:
        installed = _get_installed_distributions()
        pkgs = {_.project_name: _.version for _ in installed}
        for pkg_name, pkg_version in pkgs.items():
            include = (
                    ("duckietown" in pkg_name)
                    or ("dockertown" in pkg_name)
                    or ("dt-" in pkg_name)
                    or ("-z" in pkg_name)
                    or ("aido" in pkg_name)
            )
            if include:
                v[pkg_name] = pkg_version
    except (ImportError, AttributeError):
        logger.warning('Please update "pip" to have better debug info.')

    versions = yaml.dump(v, default_flow_style=False)
    fn = "~/shell-debug-info.txt"
    fn = os.path.expanduser(fn)
    with open(fn, "w") as f:
        f.write(versions)
    msg = f"To report a bug, please also include the contents of {fn}"
    dts_print(msg, "yellow")


def env_option(key: str, default: Any = NOTSET, true_choices: List[str] = None) -> Optional[Any]:
    if default in [NOTSET, True, False]:
        # boolean options
        true_choices = true_choices or ["1", "true", "yes"]
        default = "1" if default is True else "0"
        return os.environ.get(key, default).lower().strip() in true_choices
    else:
        # non-boolean options
        if true_choices is not None:
            raise ValueError("You cannot use 'default' and 'true_choices' at the same time.")
        return os.environ.get(key, default)
